@@ -26,6 +26,8 @@
 #define BABL_HARD_MAX_PATH_LENGTH  8
 #define BABL_MAX_NAME_LEN          1024
 
+#define BABL_TEST_ITER             4
+
 #ifndef MIN
 #define MIN(a, b) (((a) > (b)) ? (b) : (a))
 #endif
@@ -74,7 +76,7 @@ get_path_instrumentation (FishPathInstrumentation *fpi,
                           double                  *path_error);
 
 
-static long
+static inline void
 process_conversion_path (BablList   *path,
                          const void *source_buffer,
                          int         source_bpp,
@@ -99,6 +101,7 @@ _babl_fish_create_name (char       *buf,
 static int max_path_length (void);
 
 static int debug_conversions = 0;
+int _babl_instrument = 0;
 
 double _babl_legal_error (void)
 {
@@ -120,6 +123,12 @@ double _babl_legal_error (void)
   else
     debug_conversions = 0;
 
+  env = getenv ("BABL_INSTRUMENT");
+  if (env && env[0] != '\0')
+    _babl_instrument = 1;
+  else
+    _babl_instrument = 0;
+
   return error;
 }
 
@@ -210,7 +219,7 @@ get_conversion_path (PathContext *pc,
       double path_cost  = 0.0;
       double ref_cost   = 0.0;
       double path_error = 1.0;
-#if 0
+#if 1
       int    i;
       for (i = 0; i < babl_list_size (pc->current_path); i++)
         {
@@ -219,7 +228,8 @@ get_conversion_path (PathContext *pc,
 
       if (path_error - 1.0 <= legal_error )
                 /* check this before the more accurate measurement of error -
-                   to bail earlier */
+                   to bail earlier, this also leads to a stricter
+                   discarding of bad fast paths  */
 #endif
         {
           FishPathInstrumentation fpi;
@@ -423,6 +433,37 @@ alias_conversion (Babl *babl,
   return 0;
 }
 
+static void
+babl_fish_prepare_bpp (Babl *babl)
+{
+   const Babl *babl_source = babl->fish.source;
+   const Babl *babl_dest = babl->fish.destination;
+
+   switch (babl_source->instance.class_type)
+     {
+       case BABL_FORMAT:
+         babl->fish_path.source_bpp = babl_source->format.bytes_per_pixel;
+         break;
+       case BABL_TYPE:
+         babl->fish_path.source_bpp = babl_source->type.bits / 8;
+         break;
+       default:
+         babl_log ("=eeek{%i}\n", babl_source->instance.class_type - BABL_MAGIC);
+     }
+
+   switch (babl_dest->instance.class_type)
+     {
+       case BABL_FORMAT:
+         babl->fish_path.dest_bpp = babl_dest->format.bytes_per_pixel;
+         break;
+       case BABL_TYPE:
+         babl->fish_path.dest_bpp = babl_dest->type.bits / 8;
+         break;
+       default:
+         babl_log ("-eeek{%i}\n", babl_dest->instance.class_type - BABL_MAGIC);
+     }
+}
+
 
 static Babl *
 babl_fish_path2 (const Babl *source,
@@ -504,7 +545,6 @@ babl_fish_path2 (const Babl *source,
   strcpy (babl->instance.name, name);
   babl->fish.source               = source;
   babl->fish.destination          = destination;
-  babl->fish.processings          = 0;
   babl->fish.pixels               = 0;
   babl->fish.error                = BABL_MAX_COST_VALUE;
   babl->fish_path.cost            = BABL_MAX_COST_VALUE;
@@ -564,6 +604,8 @@ babl_fish_path2 (const Babl *source,
       return NULL;
     }
 
+  babl_fish_prepare_bpp (babl);
+  _babl_fish_rig_dispatch (babl);
   /* Since there is not an already registered instance by the required
    * name, inserting newly created class into database.
    */
@@ -606,76 +648,57 @@ babl_fish_path (const Babl *source,
   return babl_fish_path2 (source, destination, 0.0);
 }
 
-static long
-babl_fish_path_process (Babl       *babl,
-                        const void *source,
-                        void       *destination,
-                        long        n)
+
+static void
+babl_fish_path_process (const Babl *babl,
+                        const char *source,
+                        char       *destination,
+                        long        n,
+                        void       *data)
 {
-   const Babl *babl_source = babl->fish.source;
-   const Babl *babl_dest = babl->fish.destination;
-   int source_bpp = 0;
-   int dest_bpp = 0;
-
-   switch (babl_source->instance.class_type)
-     {
-       case BABL_FORMAT:
-         source_bpp = babl_source->format.bytes_per_pixel;
-         break;
-       case BABL_TYPE:
-         source_bpp = babl_source->type.bits / 8;
-         break;
-       default:
-         babl_log ("=eeek{%i}\n", babl_source->instance.class_type - BABL_MAGIC);
-     }
-
-   switch (babl_dest->instance.class_type)
-     {
-       case BABL_FORMAT:
-         dest_bpp = babl_dest->format.bytes_per_pixel;
-         break;
-       case BABL_TYPE:
-         dest_bpp = babl_dest->type.bits / 8;
-         break;
-       default:
-         babl_log ("-eeek{%i}\n", babl_dest->instance.class_type - BABL_MAGIC);
-     }
-
-  return process_conversion_path (babl->fish_path.conversion_list,
-                                  source,
-                                  source_bpp,
-                                  destination,
-                                  dest_bpp,
-                                  n);
+  process_conversion_path (babl->fish_path.conversion_list,
+                           source,
+                           babl->fish_path.source_bpp,
+                           destination,
+                           babl->fish_path.dest_bpp,
+                           n);
 }
 
-static long
-babl_fish_process (Babl       *babl,
-                   const void *source,
-                   void       *destination,
-                   long        n)
+static void
+babl_fish_memcpy_process (const Babl *babl,
+                          const char *source,
+                          char       *destination,
+                          long        n,
+                          void       *data)
 {
-  long ret = 0;
+  memcpy (destination, source, n * babl->fish.source->format.bytes_per_pixel);
+}
+
+void
+_babl_fish_rig_dispatch (Babl *babl)
+{
+  babl->fish.data     = (void*)&(babl->fish.data);
+
+  if (babl->fish.source == babl->fish.destination)
+    {
+      babl->fish.dispatch = babl_fish_memcpy_process;
+      return;
+    }
 
   switch (babl->class_type)
     {
       case BABL_FISH_REFERENCE:
-        if (babl->fish.source == babl->fish.destination)
-          { /* XXX: we're assuming linear buffers */
-            memcpy (destination, source, n * babl->fish.source->format.bytes_per_pixel);
-            ret = n;
-          }
-        else
-          {
-            ret = babl_fish_reference_process (babl, source, destination, n);
-          }
+        babl->fish.dispatch = babl_fish_reference_process;
         break;
 
       case BABL_FISH_SIMPLE:
         if (BABL (babl->fish_simple.conversion)->class_type == BABL_CONVERSION_LINEAR)
           {
-            ret = babl_conversion_process (BABL (babl->fish_simple.conversion),
-                                           source, destination, n);
+            /* lift out conversion from single step conversion and make it be the dispatch function
+             * itself
+             */
+            babl->fish.data = &babl->fish_simple.conversion->data;
+            babl->fish.dispatch = babl->fish_simple.conversion->dispatch;
           }
         else
           {
@@ -684,49 +707,84 @@ babl_fish_process (Babl       *babl,
         break;
 
       case BABL_FISH_PATH:
-        ret = babl_fish_path_process (babl, source, destination, n);
+        if (babl_list_size(babl->fish_path.conversion_list) == 1)
+        {
+          BablConversion *conversion = (void*)babl_list_get_first(babl->fish_path.conversion_list);
+
+          /* do same short-circuit optimization as for simple fishes */
+          babl->fish.dispatch = conversion->dispatch;
+          babl->fish.data     = &conversion->data;
+        }
+        else
+        {
+          babl->fish.dispatch = babl_fish_path_process;
+        }
+        break;
+
+      case BABL_CONVERSION:
+      case BABL_CONVERSION_LINEAR:
+      case BABL_CONVERSION_PLANE:
+      case BABL_CONVERSION_PLANAR:
+        babl_assert (0);
         break;
 
       default:
         babl_log ("NYI");
-        ret = -1;
         break;
     }
-  return ret;
+}
+
+static long
+_babl_process (const Babl *cbabl,
+               const void *source,
+               void       *destination,
+               long        n)
+{
+  Babl *babl = (void*)cbabl;
+  babl->fish.dispatch (babl, source, destination, n, *babl->fish.data);
+  if (_babl_instrument)
+    babl->fish.pixels += n;
+  return n;
 }
 
 long
-babl_process (const Babl *cbabl,
+babl_process (const Babl *babl,
               const void *source,
               void       *destination,
               long        n)
 {
-  Babl *babl = (Babl*)cbabl;
-  babl_assert (babl);
-  babl_assert (source);
-  babl_assert (destination);
-  babl_assert (BABL_IS_BABL (babl));
-  if (n == 0)
+  return _babl_process ((void*)babl, source, destination, n);
+}
+
+long
+babl_process_rows (const Babl *fish,
+                   const void *source,
+                   int         source_stride,
+                   void       *dest,
+                   int         dest_stride,
+                   long        n,
+                   int         rows)
+{
+  Babl          *babl = (Babl*)fish;
+  const uint8_t *src  = source;
+  uint8_t       *dst  = dest;
+  int            row;
+
+  babl_assert (babl && BABL_IS_BABL (babl) && source && dest);
+
+  if (n <= 0)
     return 0;
-  babl_assert (n > 0);
 
-  /* first check if it is a fish since that is our fast path */
-  if (babl->class_type >= BABL_FISH &&
-      babl->class_type <= BABL_FISH_PATH)
+  if (_babl_instrument)
+    babl->fish.pixels += n * rows;
+  for (row = 0; row < rows; row++)
     {
-      babl->fish.processings++;
-      babl->fish.pixels +=
-             babl_fish_process (babl, source, destination, n);
-      return n;
+      babl->fish.dispatch (babl, (void*)src, (void*)dst, n, *babl->fish.data);
+
+      src += source_stride;
+      dst += dest_stride;
     }
-
-  /* matches all conversion classes */
-  if (babl->class_type >= BABL_CONVERSION &&
-      babl->class_type <= BABL_CONVERSION_PLANAR)
-    return babl_conversion_process (babl, source, destination, n);
-
-  babl_fatal ("eek");
-  return -1;
+  return n * rows;
 }
 
 #include <stdint.h>
@@ -739,7 +797,7 @@ static void inline *align_16 (unsigned char *ret)
   return ret;
 }
 
-static long
+static inline void
 process_conversion_path (BablList   *path,
                          const void *source_buffer,
                          int         source_bpp,
@@ -760,13 +818,15 @@ process_conversion_path (BablList   *path,
     {
       long j;
 
-      void *temp_buffer = align_16 (alloca (MIN(n, MAX_BUFFER_SIZE) * sizeof (double) * 5 + 16));
+      void *temp_buffer = align_16 (alloca (MIN(n, MAX_BUFFER_SIZE) *
+                                    sizeof (double) * 5 + 16));
       void *temp_buffer2 = NULL;
 
       if (conversions > 2)
         {
           /* We'll need one more auxiliary buffer */
-          temp_buffer2 = align_16 (alloca (MIN(n, MAX_BUFFER_SIZE) * sizeof (double) * 5 + 16));
+          temp_buffer2 = align_16 (alloca (MIN(n, MAX_BUFFER_SIZE) *
+                                   sizeof (double) * 5 + 16));
         }
 
       for (j = 0; j < n; j+= MAX_BUFFER_SIZE)
@@ -774,15 +834,13 @@ process_conversion_path (BablList   *path,
           long c = MIN (n - j, MAX_BUFFER_SIZE);
           int i;
 
-          /* this is where the loop unrolling should happen */
           void *aux1_buffer = temp_buffer;
-          void *aux2_buffer = NULL;
-          void *swap_buffer = NULL;
-          aux2_buffer = temp_buffer2;
+          void *aux2_buffer = temp_buffer2;
 
           /* The first conversion goes from source_buffer to aux1_buffer */
           babl_conversion_process (babl_list_get_first (path),
-                                   (void*)(((unsigned char*)source_buffer) + (j * source_bpp)),
+                                   (void*)(((unsigned char*)source_buffer) +
+                                                          (j * source_bpp)),
                                    aux1_buffer,
                                    c);
 
@@ -794,21 +852,22 @@ process_conversion_path (BablList   *path,
                                        aux1_buffer,
                                        aux2_buffer,
                                        c);
-              /* Swap the auxiliary buffers */
-              swap_buffer = aux1_buffer;
-              aux1_buffer = aux2_buffer;
-              aux2_buffer = swap_buffer;
+              {
+                /* Swap the auxiliary buffers */
+                void *swap_buffer = aux1_buffer;
+                aux1_buffer = aux2_buffer;
+                aux2_buffer = swap_buffer;
+              }
             }
 
           /* The last conversion goes from aux1_buffer to destination_buffer */
           babl_conversion_process (babl_list_get_last (path),
                                    aux1_buffer,
-                                   (void*)((unsigned char*)destination_buffer + (j * dest_bpp)),
+                                   (void*)((unsigned char*)destination_buffer +
+                                                           (j * dest_bpp)),
                                    c);
         }
   }
-
-  return n;
 }
 
 static void
@@ -823,44 +882,58 @@ init_path_instrumentation (FishPathInstrumentation *fpi,
 
   if (!fpi->fmt_rgba_double)
     {
-      fpi->fmt_rgba_double = babl_format_with_space ("RGBA double",
-                                                     fmt_destination->format.space);
+      fpi->fmt_rgba_double =
+          babl_format_with_space ("RGBA double",
+                                  fmt_destination->format.space);
     }
 
   fpi->num_test_pixels = babl_get_num_path_test_pixels ();
 
-  fpi->fish_rgba_to_source      = babl_fish_reference (fpi->fmt_rgba_double,
-                                                  fmt_source);
-  fpi->fish_reference           = babl_fish_reference (fmt_source,
-                                                  fmt_destination);
-  fpi->fish_destination_to_rgba = babl_fish_reference (fmt_destination,
-                                                  fpi->fmt_rgba_double);
+  fpi->fish_rgba_to_source =
+      babl_fish_reference (fpi->fmt_rgba_double, fmt_source);
 
-  fpi->source                      = babl_calloc (fpi->num_test_pixels,
-                                             fmt_source->format.bytes_per_pixel);
-  fpi->destination                 = babl_calloc (fpi->num_test_pixels,
-                                             fmt_destination->format.bytes_per_pixel);
-  fpi->ref_destination             = babl_calloc (fpi->num_test_pixels,
-                                             fmt_destination->format.bytes_per_pixel);
-  fpi->destination_rgba_double     = babl_calloc (fpi->num_test_pixels,
-                                             fpi->fmt_rgba_double->format.bytes_per_pixel);
-  fpi->ref_destination_rgba_double = babl_calloc (fpi->num_test_pixels,
-                                             fpi->fmt_rgba_double->format.bytes_per_pixel);
+  fpi->fish_reference =
+      babl_fish_reference (fmt_source, fmt_destination);
+
+  fpi->fish_destination_to_rgba =
+      babl_fish_reference (fmt_destination, fpi->fmt_rgba_double);
+
+  fpi->source =
+      babl_calloc (fpi->num_test_pixels,
+                   fmt_source->format.bytes_per_pixel);
+
+  fpi->destination =
+      babl_calloc (fpi->num_test_pixels,
+                   fmt_destination->format.bytes_per_pixel);
+
+  fpi->ref_destination =
+      babl_calloc (fpi->num_test_pixels,
+                   fmt_destination->format.bytes_per_pixel);
+
+  fpi->destination_rgba_double =
+      babl_calloc (fpi->num_test_pixels,
+                   fpi->fmt_rgba_double->format.bytes_per_pixel);
+
+  fpi->ref_destination_rgba_double =
+      babl_calloc (fpi->num_test_pixels,
+                   fpi->fmt_rgba_double->format.bytes_per_pixel);
 
   /* create sourcebuffer from testbuffer in the correct format */
-  babl_process (fpi->fish_rgba_to_source,
-                test_pixels, fpi->source, fpi->num_test_pixels);
+  _babl_process (fpi->fish_rgba_to_source,
+                 test_pixels, fpi->source,fpi->num_test_pixels);
 
   /* calculate the reference buffer of how it should be */
   ticks_start = babl_ticks ();
-  babl_process (fpi->fish_reference,
-                fpi->source, fpi->ref_destination, fpi->num_test_pixels);
+  _babl_process (fpi->fish_reference,
+                 fpi->source, fpi->ref_destination,
+                 fpi->num_test_pixels);
   ticks_end = babl_ticks ();
-  fpi->reference_cost = ticks_end - ticks_start;
+  fpi->reference_cost = (ticks_end - ticks_start) * BABL_TEST_ITER;
 
   /* transform the reference destination buffer to RGBA */
-  babl_process (fpi->fish_destination_to_rgba,
-                fpi->ref_destination, fpi->ref_destination_rgba_double, fpi->num_test_pixels);
+  _babl_process (fpi->fish_destination_to_rgba,
+                 fpi->ref_destination, fpi->ref_destination_rgba_double,
+                 fpi->num_test_pixels);
 }
 
 static void
@@ -916,7 +989,8 @@ get_path_instrumentation (FishPathInstrumentation *fpi,
         dest_bpp = babl_destination->type.bits / 8;
         break;
       default:
-        babl_log ("-eeek{%i}\n", babl_destination->instance.class_type - BABL_MAGIC);
+        babl_log ("-eeek{%i}\n",
+                  babl_destination->instance.class_type - BABL_MAGIC);
      }
 
   if (!fpi->init_instrumentation_done)
@@ -930,19 +1004,22 @@ get_path_instrumentation (FishPathInstrumentation *fpi,
 
   /* calculate this path's view of what the result should be */
   ticks_start = babl_ticks ();
-  process_conversion_path (path, fpi->source, source_bpp, fpi->destination, dest_bpp, fpi->num_test_pixels);
+  for (int i = 0; i < BABL_TEST_ITER; i ++)
+  process_conversion_path (path, fpi->source, source_bpp, fpi->destination,
+                           dest_bpp, fpi->num_test_pixels);
   ticks_end = babl_ticks ();
-  *path_cost = ticks_end - ticks_start;
+  *path_cost = (ticks_end - ticks_start);
 
   /* transform the reference and the actual destination buffers to RGBA
    * for comparison with each other
    */
-  babl_process (fpi->fish_destination_to_rgba,
-                fpi->destination, fpi->destination_rgba_double, fpi->num_test_pixels);
+  _babl_process (fpi->fish_destination_to_rgba,
+                 fpi->destination, fpi->destination_rgba_double,
+                 fpi->num_test_pixels);
 
   *path_error = babl_rel_avg_error (fpi->destination_rgba_double,
                                     fpi->ref_destination_rgba_double,
-                                     fpi->num_test_pixels * 4);
+                                    fpi->num_test_pixels * 4);
 
   *ref_cost = fpi->reference_cost;
 }
