@@ -66,10 +66,6 @@ Babl *   babl_conversion_find           (const void     *source,
                                          const void     *destination);
 double   babl_conversion_error          (BablConversion *conversion);
 long     babl_conversion_cost           (BablConversion *conversion);
-long     babl_conversion_process        (const Babl     *conversion,
-                                         const char     *source,
-                                         char           *destination,
-                                         long            n);
 
 Babl   * babl_extension_base            (void);
 
@@ -79,10 +75,11 @@ void     babl_set_extender              (Babl           *new_extender);
 Babl   * babl_extension_quiet_log       (void);
 void     babl_extension_deinit          (void);
 
-long     babl_fish_reference_process    (const Babl *babl,
+void     babl_fish_reference_process    (const Babl *babl,
                                          const char *source,
                                          char       *destination,
-                                         long        n);
+                                         long        n,
+                                         void       *data); // data is ignored
 
 Babl   * babl_fish_reference            (const Babl     *source,
                                          const Babl     *destination);
@@ -486,4 +483,20 @@ void babl_space_to_xyz   (const Babl *space, const double *rgb, double *xyz);
  */
 void babl_space_from_xyz (const Babl *space, const double *xyz, double *rgb);
 
+extern int _babl_instrument;
+
+static inline void
+babl_conversion_process (const Babl *babl,
+                         const char *source,
+                         char       *destination,
+                         long        n)
+{
+  BablConversion *conversion = (BablConversion *) babl;
+  if (_babl_instrument)
+    conversion->pixels += n;
+  conversion->dispatch (babl, source, destination, n, conversion->data);
+}
+
+void _babl_fish_rig_dispatch (Babl *babl);
+
 #endif
